#ifndef FLY_ICON_THEME
#define FLY_ICON_THEME

#include <fly/flytaggedbool.h>

using Display = struct _XDisplay;

#include <stdlib.h>

//#if defined(__linux__) || defined(_linux_) || defined(linux)
//#include <linux/limits.h> //for PATH_MAX, NAME_MAX
//#else
#include <limits.h>
//#endif

#include <vector>
#include <string>

#define MAX_INHERITS_NUM          2
#define CONTEXT_SIZE             30

//helper class for FlyIconTheme class
class FlyIconThemeSubdirPrivate;
class FlyIconThemeSubdir
{
public:
    FlyIconThemeSubdir(std::string path);
    FlyIconThemeSubdir(const FlyIconThemeSubdir &other) = delete;
    FlyIconThemeSubdir(FlyIconThemeSubdir &&other) noexcept;
    ~FlyIconThemeSubdir();

    FlyIconThemeSubdir &operator=(const FlyIconThemeSubdir &other) = delete;
    FlyIconThemeSubdir &operator=(FlyIconThemeSubdir &&other) noexcept;

    // set/get standard icon theme dir attributes
    int   size() const;
    void  setSize(int sz);

    int   maxSize() const;
    void  setMaxSize(int msz);

    int   minSize() const;
    void  setMinSize(int msz);

    int   threshold() const;
    void  setThreshold(int t);

    const std::string &context() const;
    void setContext(std::string context);

    const std::string &type() const;
    void setType(std::string type);

    const std::string &path() const;
    void setPath(std::string path);

    // find full icon path by icon name (without extension!)
    std::string findIcon(const char *themePath, const char *iconName, const int iconSize,
                         const char *iconContext, const std::vector<std::string>& spreadPaths) const;
    bool matchesSize(int size) const;

private:
    FlyIconThemeSubdirPrivate * p;
};

class FlyIconThemePrivate;
class FlyIconTheme
{
public:
    enum SearchIconNameType {
    SearchMainIcon    = 0x1, //search main icon only
    SearchGenericIcon = 0x2, //search generic icon only
    SearchAnyIcon     = SearchMainIcon | SearchGenericIcon //seach main and if nto found then generic icon
    };

    FlyIconTheme(std::string name = std::string(), std::string parentName = std::string(), bool forceLoad = false);
    FlyIconTheme(const FlyIconTheme &other) = delete;
    FlyIconTheme(FlyIconTheme &&other) noexcept;
    ~FlyIconTheme();

    FlyIconTheme &operator=(const FlyIconTheme &other) = delete;
    FlyIconTheme &operator=(FlyIconTheme &&other) noexcept;

    // create static object of current theme and return pointer to it
    // this method checks if icon theme was changed and create new one if so
    // it's not thread-safe
    using ForceLoad = Fly::tagged_bool<class ForceLoad_tag>;
    static FlyIconTheme *current(ForceLoad forceLoad = ForceLoad{false});
    static FlyIconTheme *currentTray(ForceLoad forceLoad = ForceLoad{false}, std::string forColor="");

    // create static object of current theme and return pointer to it
    // theme created by this method is not updated during the execution of the program
    // it's thread-safe
    static FlyIconTheme *instance();

    const std::string &path() const; // returns full path to index.theme file
    void setPath(std::string path);  // set     full path to index.theme file

    const std::string &themeName() const; // get theme internal name
    void setThemeName(std::string name);  // set theme internal name

    bool  load(); //reload index.theme file
    bool  save(); //save main attributes (see below) in index.theme file
    bool  isLoaded() const; //was it loaded at least once or not

    // set/get standard icon theme attributes

    const std::string &name() const; // get theme name
    void setName(std::string name);  // set theme name

    const std::string &localName() const; // get theme name in local language
    void setLocalName(std::string name);  // set theme name in local language
    
    const std::string &comment() const;   // get theme description
    void setComment(std::string comment); // set theme description

    const std::string &localComment() const;  // get theme description in local language
    void setLocalComment(std::string comment); // set theme description in local language

    bool hidden() const; // true if theme is hidden (may be we sholdn't show it)
    void setHidden(bool);

    const std::string &exampleIcon() const;   // picture as theme example (usually icon name without extension)
    void setExampleIcon(std::string example); // picture as theme example (usually icon name without extension)

    const std::vector<std::string> &inheritedNames() const; // get inherited themes names
    void setInheritedNames(std::vector<std::string> names); // set inherited themes names

    const std::vector<FlyIconThemeSubdir> &subdirs() const;    // get subdirs
    void setSubdirs(std::vector<FlyIconThemeSubdir> subdirs); // set subdirs

    // main API for apps: ---------------------------------------------

    // get icon theme path and description file name
    std::string findThemePath(const std::string &internalName);

    // find full icon path by icon short name (without extension!), by size and by context
    // contextIn - may be NULL or "" (then guessing will be by getIconContextAndPath())
    std::string getIconPath(const char *iconNameIn, int iconSizeIn, const char *contextIn = nullptr);

    // guess icon context by name and get full path if required
    // ctx_ret - will be set to icon context
    std::string getIconContextAndPath(const char *iconName, const int iconSize, std::string *ctx_ret);

    // get full icon  path for mime type name and size
    // xdg_mime_get_icon() used first
    std::string getMimeIconPath(const char *mimeNameIn, int iconSizeIn);

    // get mime icon name only, with check for full path (default size 32)
    std::string getMimeIconName(const char *mimeNameIn);

    // get main and-or generic icon name from shared-mime-info database, depends on shared-mime-info completely
    std::string findMimeIconNameXDG(const char *mimeNameIn, SearchIconNameType searchType) const;

    // get main icon name from shared-mime-info database, depends on shared-mime-info completely
    std::string getMimeIconNameXDG(const char *mimeNameIn) const;

    // get generic icon name from shared-mime-info database, depends on shared-mime-info completely
    std::string getMimeGenericIconNameXDG(const char *mimeNameIn) const;

    // just replace '/' with '-', i.e. text/x-makefile -> text-x-makefile
    std::string getMimeIconNameSimple(const char *mimeNameIn) const;

    // fullPath- has to be non NULL field provided by caller
    std::string getURLIconPath(const char *url, const int iconSize);

    std::string getURLIconName(const char *url, std::string *contextOut = nullptr);

    std::string getURLIcon(const char  *url,
                           const int   iconSize,
                           std::string *nameOut,
                           std::string *ctxOut);

    // if iconName is full file name or path - check it in predefined directories
    static std::string CheckFullPath(const char *iconName, std::string *ctx_ret = nullptr);

    //get colon separated dirs to search themes and icons
    static std::vector<std::string> GetThemesSearchDirs();

    //fill possible themes list by pattern ("index.* or "cursors")
    static std::vector<std::string> GetThemesIndexFiles();

    //get name of current icon theme from ~/.fly/theme/current.themerc or some default
    static std::string getCurrentThemeName();
    static std::string getCurrentTrayThemeName(std::string forColor="");


    static std::vector<std::string> getThemesNames();

    static std::vector<std::string> getCursorThemePaths();
    static std::vector<std::string> getCursorThemeNames();
    static std::string getCurrentCursorThemeName(Display *dpy = nullptr); // get 'Xcursor.theme: ' from ~/.Xresources
    static bool setCurrentCursorThemeName(const char *themeName);         // set 'Xcursor.theme: ' in   ~/.Xresources

    //returns suitable emblem icon size to make overlay with main icon
    static int GetIconEmblemSize(int iconSize/*, unsigned int dpi=96*/);

    //return one of standard sizes to fit (less) itemSize
    static int GetStandardIconSize(int itemSize);

    //recommended icon size
    static int GetIconSizeForDPI(int dpi, int inc);

    static int GetFrameTitleSizeForDPI(int dpi, int def);
    static int GetFrameBorderSizeForDPI(int dpi, int def);

    // ---------------------------------------------------------------
private:
    FlyIconThemePrivate * p;
};

#endif

