#ifndef FLY_STRING_UTILS_H
#define FLY_STRING_UTILS_H

#include <vector>
#include <string>

#include <string.h>

namespace Fly
{
//recoding (don't forget to cast 'in' and 'out' to (unsigned char *) in these function calls)
bool isValidUtf8(const unsigned char *);
bool isEscapedUtf8(const unsigned char *buffer, unsigned char *utf8buf_out=nullptr);

bool needDoublequotes(const char *str);
std::string quoteArg(const char *s);
inline std::string quoteArg(const std::string &s) { return quoteArg(s.c_str()); }

inline std::string makeString(const char *s) { return s ? s : std::string(); }
std::vector<std::string> split(const char *str, const char *delim);
std::string join(const std::vector<std::string> &strings, const char *sep);
char *newStringCopy(const std::string &str); // the duplicated string must be freed with delete[]
bool startsWith(const char *str, const char *beginning);
bool startsWith(const std::string &str, const char *beginning);
bool startsWith(const std::string &str, const std::string &beginning);
bool startsWith(const char *str, const std::string &beginning);
bool endsWith(const char *str, const char *ending);
bool endsWith(const std::string &str, const char *ending);
bool endsWith(const std::string &str, const std::string &ending);
inline bool equal(const char *s1, const char *s2) { return strcmp(s1, s2) == 0; }
inline bool ciEqual(const std::string &s1, const std::string &s2) { return strcasecmp(s1.c_str(), s2.c_str()) == 0; }
inline bool ciEqual(const std::string &s1, const char *s2) { return strcasecmp(s1.c_str(), s2) == 0; }
inline bool contains(const char *str, const char *substr) { return strstr(str, substr); }
inline bool contains(const char *str, const std::string &substr) { return contains(str, substr.c_str()); }
inline bool contains(const std::string &str, const char *substr) { return contains(str.c_str(), substr); }
inline bool contains(const std::string &str, const std::string &substr) { return contains(str.c_str(), substr.c_str()); }
inline bool ciContains(const char *str, const char *substr) { return strcasestr(str, substr); }
inline bool ciContains(const char *str, const std::string &substr) { return ciContains(str, substr.c_str()); }
inline bool ciContains(const std::string &str, const char *substr) { return ciContains(str.c_str(), substr); }
inline bool ciContains(const std::string &str, const std::string &substr) { return ciContains(str.c_str(), substr.c_str()); }
std::string baseName(const std::string &path);
std::string completeBaseName(const std::string &path);
bool cutLastSection(std::string &s, char sep);
std::string lastSection(const std::string &s, char sep);
std::string section(const std::string &s, char sep, int sect);

//encode/decode trashed file paths (rfc1738)
std::string encodedUrl(const char *s);
std::string encodedUrl(const std::string &s);
std::string decodedUrl(const char *s);
std::string decodedUrl(const std::string &s);

namespace Impl
{
    inline void append(std::string &) {}

    template <typename... Args>
    inline void append(std::string &str, char c, Args &&... args);

    template <typename... Args>
    inline void append(std::string &str, const std::string &s, Args &&... args);

    template <typename... Args>
    inline void append(std::string &str, const char *s, Args &&... args);

    template <typename Arg, typename... Args, typename = typename std::enable_if<std::is_arithmetic<typename std::decay<Arg>::type>::value &&
                                                                                 !std::is_same<Arg, char>::value>::type>
    inline void append(std::string &str, Arg arg, Args &&... args)
    {
        str += std::to_string(arg);
        append(str, std::forward<Args>(args)...);
    }

    template <typename... Args>
    inline void append(std::string &str, char c, Args &&... args)
    {
        str += c;
        append(str, std::forward<Args>(args)...);
    }

    template <typename... Args>
    inline void append(std::string &str, const std::string &s, Args &&... args)
    {
        str += s;
        append(str, std::forward<Args>(args)...);
    }

    template <typename... Args>
    inline void append(std::string &str, const char *s, Args &&... args)
    {
        if (s)
            str += s;
        append(str, std::forward<Args>(args)...);
    }
}

template <typename... Args>
inline std::string concat(Args &&... args)
{
    std::string result;
    Impl::append(result, args...);
    return result;
}

}

#endif
